# -*- coding: utf-8 -*-

'''
Created 17-Dec-2024
Updated 29-Dec-2024
'''

from MSPyBentley import *
from MSPyDgnPlatform import *
from MSPyMstnPlatform import *
# We use a Counter object to count the level ID found on each graphical element
from collections import Counter

def getElementDescription (element, stringLength: int = 32)->str:
    '''
    Get the human-readable description of an element.
    '''
    description = WString()
    if isinstance (element, ElementHandle):
        handler = element.GetHandler()
        handler.GetDescription (element, description, stringLength)
        
    if isinstance (element, PersistentElementRef):
        eh = ElementHandle (element, dgnModel)
        handler = eh.GetHandler()       
        handler.GetDescription (eh, description, stringLength)
        
    return str(description)

'''
Get the level ID of a named level in the specified DGN file
    lvlName : str              Level Name
    returns lvlId   : int      Level ID
'''
def getLevelByName (levelName, dgnModel = ISessionMgr.ActiveDgnModelRef)->int:
    levelId: int = 0
    levelCache = dgnModel.GetDgnFile().GetLevelCache()
    for level in levelCache:
        name = level.GetName()
        #   Case-insensitive text comparison
        if name.casefold() == levelName.casefold():
            levelId = level.GetLevelId()
            MessageCenter.ShowDebugMessage(f"getLevelByName found level '{levelName}'", f"getLevelByName found level '{levelName}' LevelId={levelId}", False)
            return levelId
    MessageCenter.ShowErrorMessage(f"getLevelByName failed to find level '{levelName}'", f"getLevelByName failed to find level '{levelName}'", False)
    return 0
    
# A Python tuple of DGN element handlers    
ComplexElementHandlers = (NormalCellHeaderHandler, GroupedHoleHandler, ) 

# ExposeChildrenReason instance required by the ChildElemIter() constructor.
# See: https://bentleysystems.service-now.com/community?id=community_question&sys_id=3950e8461b6612147171ff7e034bcb58
ExposeChildrenCount = ExposeChildrenReason(100)
ExposeChildrenQuery = ExposeChildrenReason(200)
ExposeChildrenEdit  = ExposeChildrenReason(300)

# see collections.Counter:
# https://docs.python.org/3/library/collections.html#collections.Counter
def CollectLevels (elemRef: PersistentElementRef, dgnModel: DgnModel, levelCounter: Counter):
    """Get level(s) from an element, and insert into a Counter keyed by level ID.
    This function calls itself recursively to investigate complex elements such as cells."""

    eh = ElementHandle (elemRef, dgnModel)
    handler = eh.GetHandler()
    description = getElementDescription(eh)
    MessageCenter.ShowDebugMessage(f"CollectLevels found {description} ComplexElementHandler={handler in ComplexElementHandlers}", 
        f"CollectLevels found {description}  in ComplexElementHandlers={handler in ComplexElementHandlers}", False)
    #   If this is a complex element  (e.g. a cell or grouped hole)
    if isinstance (handler, ComplexElementHandlers):
        component = ChildElemIter(eh, ExposeChildrenCount)
        while component.IsValid():
            handler = component.GetHandler()
            if isinstance (handler, ComplexElementHandlers):
                CollectLevels (component.GetElementRef(), dgnModel, levelCounter)
            else:
                levelId = component.GetElement().ehdr.level
                levelCounter.update ((levelId, ))
                
            component = component.ToNext()
    else:
        levelId = eh.GetElement().ehdr.level
        count = levelCounter[levelId]
        MessageCenter.ShowDebugMessage(f"CollectLevels found {count} level ID {levelId}", f"CollectLevels found level ID {levelId}", False)
        levelCounter.update ((levelId, ))
        
class DgnElementLevelCounter:
    """A class to count DGN elements on one or more levels in a DGN model.
    
    For example, count elements on level 'Areas'.    
    By default, collects elements from the active model.
    For each element, it adds its level ID to a Counter object.    
    """
    
    def __init__(self, dgnModel: DgnModel = ISessionMgr.ActiveDgnModelRef.GetDgnModel()):             
        self._dgnModel = dgnModel
        self._elementCount = 0
        self._levelCounter = Counter()
        
    def EnumerateElements(self)->int:
        nElements = 0
        for elemRef in self._dgnModel.GetGraphicElements(): 
            nElements += 1
            CollectLevels (elemRef, self._dgnModel, self._levelCounter)
            
        return nElements

    @property
    def LevelCounter (self):
        # Get the level counter object
        return self._levelCounter
        
    @property
    def DgnModelName (self):
        # Get the name of the DGN model that contains elements to be counted
        if (self._dgnModel is None): 
            return "unknown"
        else:
            return self._dgnModel.GetModelName()

    def __str__(self):
        return f"Element count {self._elementCount} in model '{self.DgnModelName()}'"
        
    def __repr__(self):
        return 'CountElements(Level ID List)'
                
#main
if __name__ == "__main__":
    MessageCenter.ShowDebugMessage("DgnElementCounter", "Count element on a named level", False)
    # Construct a new Counter object
    counter = DgnElementLevelCounter()
    nElements = counter.EnumerateElements()
    MessageCenter.ShowDebugMessage(f"Found {nElements} elements", f"Found {nElements} elements in total in model '{counter.DgnModelName}'", False)
    # You can get the total number of levels found on all elements
    nLevels = counter.LevelCounter.total()
    MessageCenter.ShowDebugMessage(f"Found {nLevels} levels in total", f"Found {nLevels} levels in total in model '{counter.DgnModelName}'", False)
    
    userLevelName = "Areas Type 2"
    levelId = getLevelByName(userLevelName)
    if 0 != levelId:
        nLevels = counter.LevelCounter[levelId]
        MessageCenter.ShowDebugMessage(f"Found {nLevels} elements on level '{userLevelName}'", f"Found {nLevels} elements on level '{userLevelName}' in model '{counter.DgnModelName}'", False)
        
    """
    counter = DgnElementLevelCounter()
    userLevelName = "I don't exist"
    levelId = getLevelByName(userLevelName)  
    if 0 != levelId:
        nLevels = counter.LevelCounter[levelId]
        MessageCenter.ShowDebugMessage(f"Found {nLevels} levels on level '{userLevelName}'", f"Found {nLevels} levels on level '{userLevelName}' in model '{counter.DgnModelName}'", False)
    
    userLevelName = "Areas"
    levelId = getLevelByName(userLevelName)
    if 0 != levelId:
        nLevels = counter.LevelCounter[levelId]
        MessageCenter.ShowDebugMessage(f"Found {nLevels} elements on level '{userLevelName}'", f"Found {nLevels} elements on level '{userLevelName}' in model '{counter.DgnModelName}'", False)
    """